'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});

var _fs = require('fs');

var _fs2 = _interopRequireDefault(_fs);

var _espree = require('espree');

var _espree2 = _interopRequireDefault(_espree);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const EspreeConfig = {
    // attach range information to each node
    range: true,
    // attach line/column location information to each node
    loc: true,
    // create a top-level comments array containing all comments
    comment: true,
    // attach comments to the closest relevant node as leadingComments and
    // trailingComments
    attachComment: true,
    // create a top-level tokens array containing all tokens
    tokens: true,
    // specify the language version (3, 5, 6, or 7, default is 5)
    ecmaVersion: 5,
    // specify which type of script you're parsing (script or module, default is script)
    sourceType: 'module',
    // specify additional language features
    ecmaFeatures: {
        // enable JSX parsing
        jsx: true,
        // enable return in global scope
        globalReturn: true,
        // enable implied strict mode (if ecmaVersion >= 5)
        impliedStrict: true,
        // allow experimental object rest/spread
        experimentalObjectRestSpread: true,
        // allow let and const declarations
        blockBindings: true
    }
};

// Leaving this code here in case support for assignment variables are needed.
// Since we are only supporting "const" for now, this isn't needed.
const ExtractAssignmentConstants = (node, store) => {
    if (!node) {
        throw new Error('Something went wrong!');
    }
    if (node.type !== 'AssignmentExpression') {
        return node.end;
    }
    var _node$left = node.left;
    const name = _node$left.name;
    const start = _node$left.start;
    const end = _node$left.end;

    store.push({ name, start, end });
    ExtractAssignmentConstants(node.right, store);
};

const ExtractGlobals = astBody => {
    const globals = {
        imports: {
            start: null,
            end: null,
            nodes: []
        },
        constants: {
            start: null,
            end: null,
            nodes: []
        },
        exports: {
            start: null,
            end: null,
            nodes: []
        }
    };

    astBody.forEach(node => {
        if (node.type === 'ImportDeclaration') {
            // handle imports
            const start = node.start;
            const end = node.end;
            const specifiers = node.specifiers;

            if (globals.imports.start === null) {
                globals.imports.start = start;
            }
            globals.imports.end = end;
            specifiers.forEach(node => {
                var _node$local = node.local;
                const name = _node$local.name;
                const start = _node$local.start;
                const end = _node$local.end;

                globals.imports.nodes.push({ name, start, end });
            });
        } else if (node.type === 'ExportAllDeclaration' || // handle exports
        node.type === 'ExportDefaultDeclaration' || node.type === 'ExportNamedDeclaration') {
            if (globals.exports.start === null) {
                globals.exports.start = node.start;
            }
            if (node.type === 'ExportDefaultDeclaration') {
                var _node$declaration = node.declaration;
                const name = _node$declaration.name;
                const start = _node$declaration.start;
                const end = _node$declaration.end;

                globals.exports.nodes.push({ name, start, end });
            } else if (node.type === 'ExportNamedDeclaration') {
                if (node.declaration && node.declaration.declarations[0]) {
                    var _node$declaration$dec = node.declaration.declarations[0];
                    const id = _node$declaration$dec.id;
                    const init = _node$declaration$dec.init;
                    const name = id.name;
                    const start = id.start;
                    const end = id.end;

                    globals.exports.nodes.push({ name, start, end });
                    ExtractAssignmentConstants(init, globals.exports.nodes);
                } else {
                    node.specifiers.forEach(node => {
                        var _node$local2 = node.local;
                        const name = _node$local2.name;
                        const start = _node$local2.start;
                        const end = _node$local2.end;

                        globals.exports.nodes.push({ name, start, end });
                    });
                }
            }
            globals.exports.end = node.end;
        } else if (node.type === 'VariableDeclaration' && // handle constants
        node.kind === 'const') {
            if (globals.constants.start === null) {
                globals.constants.start = node.start;
            }
            node.declarations.forEach(node => {
                const id = node.id;
                const init = node.init;
                const name = id.name;
                const start = id.start;
                const end = id.end;

                const constantNodes = globals.constants.nodes;
                const terminal = ExtractAssignmentConstants(init, constantNodes);
                constantNodes.push({ name, start, end, terminal });
            });
            globals.constants.end = node.end;
        }
    });
    return globals;
};

const Parse = code => {
    if (typeof code !== 'string') {
        throw new Error('Please pass in a valid code string!');
    }
    const ast = _espree2.default.parse(code, EspreeConfig);
    return ExtractGlobals(ast.body);
};

exports.default = Parse;